"""Generate a report for each gene with variant count per impact and pathogenicity"""

import enum
import gzip
import json
import sys
import typing

import tqdm

from clinvar_data import models
from clinvar_data.cattrs_helpers import CONVERTER


@enum.unique
class Impact(enum.Enum):
    """SO terms for gene impact"""

    THREE_PRIME_UTR_VARIANT = "3_prime_UTR_variant"
    FIVE_PRIME_UTR_VARIANT = "5_prime_UTR_variant"
    DOWNSTREAM_TRANSCRIPT_VARIANT = "downstream_gene_variant"
    FRAMESHIFT_VARIANT = "frameshift_variant"
    INFRAME_INDEL = "inframe_indel"
    START_LOST = "start_lost"
    INTRON_VARIANT = "intron_variant"
    MISSENSE_VARIANT = "missense_variant"
    NON_CODING_TRANSCRIPT_VARIANT = "non_coding_transcript_variant"
    STOP_GAINED = "stop_gained"
    NO_SEQUENCE_ALTERATION = "no_sequence_alteration"
    SPLICE_ACCEPTOR_VARIANT = "splice_acceptor_variant"
    SPLICE_DONOR_VARIANT = "splice_donor_variant"
    STOP_LOST = "stop_lost"
    SYNONYMOUS_VARIANT = "synonymous_variant"
    UPSTREAM_TRANSCRIPT_VARIANT = "upstream_gene_variant"


#: Mapping from strings used in ClinVar XML to ``IMPACT``.
GENE_IMPACT_MAP = {
    "3 prime UTR variant": Impact.THREE_PRIME_UTR_VARIANT,
    "5 prime UTR variant": Impact.FIVE_PRIME_UTR_VARIANT,
    "downstream transcript variant": Impact.DOWNSTREAM_TRANSCRIPT_VARIANT,
    "frameshift variant": Impact.FRAMESHIFT_VARIANT,
    "genic downstream transcript variant": Impact.DOWNSTREAM_TRANSCRIPT_VARIANT,
    "genic upstream transcript variant": Impact.UPSTREAM_TRANSCRIPT_VARIANT,
    "inframe_deletion": Impact.INFRAME_INDEL,
    "inframe_indel": Impact.INFRAME_INDEL,
    "inframe_insertion": Impact.INFRAME_INDEL,
    "initiatior codon variant": Impact.START_LOST,
    "intron variant": Impact.INTRON_VARIANT,
    "missense variant": Impact.MISSENSE_VARIANT,
    "non-coding transcript variant": Impact.NON_CODING_TRANSCRIPT_VARIANT,
    "nonsense": Impact.STOP_GAINED,
    "no sequence alteration": Impact.NO_SEQUENCE_ALTERATION,
    "splice acceptor variant": Impact.SPLICE_ACCEPTOR_VARIANT,
    "splice donor variant": Impact.SPLICE_DONOR_VARIANT,
    "stop lost": Impact.STOP_LOST,
    "synonymous variant": Impact.SYNONYMOUS_VARIANT,
    "upstream transcript variant": Impact.UPSTREAM_TRANSCRIPT_VARIANT,
}

#: ACMG clinical significance values
ACMG_CLINSIGS = (
    models.ClinicalSignificanceDescription.BENIGN,
    models.ClinicalSignificanceDescription.LIKELY_BENIGN,
    models.ClinicalSignificanceDescription.UNCERTAIN_SIGNIFICANCE,
    models.ClinicalSignificanceDescription.LIKELY_PATHOGENIC,
    models.ClinicalSignificanceDescription.PATHOGENIC,
)


def zero_counts() -> typing.Dict[typing.Tuple[Impact, models.ClinicalSignificanceDescription], int]:
    return {(impact, pathogenicity): 0 for impact in Impact for pathogenicity in ACMG_CLINSIGS}


def write_report(counts: dict, path_output: str):
    """Write report to output."""

    if path_output.endswith(".gz"):
        outputf = gzip.open(path_output, "wt")
    else:
        outputf = open(path_output, "wt")

    with outputf:
        for hgnc, counts in sorted(counts.items(), key=lambda x: int(x[0][5:])):
            counts_out = {}
            for impact in Impact:
                arr = []
                for patho in ACMG_CLINSIGS:
                    arr.append(counts[(impact, patho)])
                if sum(arr) > 0:
                    counts_out[impact.value] = arr
            print(
                json.dumps({"hgnc": hgnc, "counts": counts_out}),
                file=outputf,
            )


def generate_counts(path_input: str) -> dict:
    """Count occurrences of each impact for each gene."""
    counts = {}

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    with inputf:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            dict_value = json.loads(line)
            clinvar_set = CONVERTER.structure(dict_value, models.ClinVarSet)

            pathogenicity = (
                clinvar_set.reference_clinvar_assertion.clinical_significance.description
            )
            if not pathogenicity or not pathogenicity.is_canonical_acmg:
                continue  # skip not ACMG 1-5

            if not clinvar_set.reference_clinvar_assertion.measures:
                continue

            csq = None
            for measure in clinvar_set.reference_clinvar_assertion.measures.measures:
                for attribute in measure.attributes:
                    if attribute.type == models.MeasureAttributeType.MOLECULAR_CONSEQUENCE:
                        csq = attribute.value
                        break
            if not csq:
                continue  # skip, no molecular consequence
            elif csq not in GENE_IMPACT_MAP:
                print(f"WARNING: unknown molecular consequence {csq}", file=sys.stderr)
                continue

            hgnc = None
            for measure in clinvar_set.reference_clinvar_assertion.measures.measures:
                for measure_relationship in measure.measure_relationship:
                    for xref in measure_relationship.xrefs:
                        if xref.db == "HGNC":
                            hgnc = xref.id
                            break
            if not hgnc:
                continue  # skip, no HGNC ID

            if hgnc not in counts:
                counts[hgnc] = zero_counts()
            counts[hgnc][(GENE_IMPACT_MAP[csq], pathogenicity)] += 1
    return counts


def run_report(path_input: str, path_output: str):
    """Generate the report from the given input to output path."""
    counts = generate_counts(path_input)
    write_report(counts, path_output)



"""Helper code for cattrs."""
import datetime

import cattrs
from dateutil.parser import parse as parse_datetime

CONVERTER = cattrs.Converter()
CONVERTER.register_unstructure_hook(datetime.date, lambda d: d.strftime("%Y-%m-%d"))
CONVERTER.register_unstructure_hook(datetime.datetime, lambda d: d.strftime("%Y-%m-%d %H:%M:%S"))
CONVERTER.register_structure_hook(datetime.date, lambda s, _: parse_datetime(s).date())
CONVERTER.register_structure_hook(datetime.datetime, lambda s, _: parse_datetime(s))



"""Data structures for internal representation of submission payload."""

import typing

import attrs

from clinvar_api import msg

# Import and re-use the enums. This makes coupling tighter than it would have to be but the
# trade-off of copy-pasting them with no influence on the interal record API makes that not
# worth it.
from clinvar_api.msg.sub_payload import (
    AffectedStatus,
    AlleleOrigin,
    Assembly,
    Chromosome,
    CitationDb,
    ClinicalFeaturesAffectedStatus,
    ClinicalFeaturesDb,
    ClinicalSignificanceDescription,
    CollectionMethod,
    ConditionDb,
    ModeOfInheritance,
    RecordStatus,
    ReleaseStatus,
    StructVarMethodType,
    VariantType,
)


@attrs.define(frozen=True)
class SubmissionClinvarDeletionAccessionSet:
    accession: str
    reason: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionClinvarDeletionAccessionSet:
        return msg.SubmissionClinvarDeletionAccessionSet(
            accession=self.accession,
            reason=self.reason,
        )


@attrs.define(frozen=True)
class SubmissionClinvarDeletion:
    accession_set: typing.List[SubmissionClinvarDeletionAccessionSet]

    def to_msg(self) -> msg.SubmissionClinvarDeletion:
        return msg.SubmissionClinvarDeletion(
            accessionSet=[msg_accession_set.to_msg() for msg_accession_set in self.accession_set],
        )


@attrs.define(frozen=True)
class SubmissionChromosomeCoordinates:
    accession: typing.Optional[str] = None
    alternate_allele: typing.Optional[str] = None
    assembly: typing.Optional[Assembly] = None
    chromosome: typing.Optional[Chromosome] = None
    inner_start: typing.Optional[int] = None
    inner_stop: typing.Optional[int] = None
    outer_start: typing.Optional[int] = None
    outer_stop: typing.Optional[int] = None
    reference_allele: typing.Optional[str] = None
    start: typing.Optional[int] = None
    stop: typing.Optional[int] = None
    variant_length: typing.Optional[int] = None

    def to_msg(self) -> msg.SubmissionChromosomeCoordinates:
        return msg.SubmissionChromosomeCoordinates(
            accession=self.accession,
            assembly=self.assembly,
            alternateAllele=self.alternate_allele,
            chromosome=self.chromosome,
            innerStart=self.inner_start,
            innerStop=self.inner_stop,
            outerStart=self.outer_start,
            outerStop=self.outer_stop,
            referenceAllele=self.reference_allele,
            start=self.start,
            stop=self.stop,
            variantLength=self.variant_length,
        )


@attrs.define(frozen=True)
class SubmissionVariantGene:
    id: typing.Optional[int] = None
    symbol: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionVariantGene:
        return msg.SubmissionVariantGene(
            id=self.id,
            symbol=self.symbol,
        )


@attrs.define(frozen=True)
class SubmissionVariant:
    chromosome_coordinates: typing.Optional[SubmissionChromosomeCoordinates] = None
    copy_number: typing.Optional[str] = None
    gene: typing.Optional[typing.List[SubmissionVariantGene]] = None
    hgvs: typing.Optional[str] = None
    reference_copy_number: typing.Optional[int] = None
    variant_type: typing.Optional[VariantType] = None

    def to_msg(self) -> msg.SubmissionVariant:
        chromosome_coordinates = None
        if self.chromosome_coordinates:
            chromosome_coordinates = self.chromosome_coordinates.to_msg()
        gene = None
        if self.gene:
            gene = [msg_gene.to_msg() for msg_gene in self.gene]
        return msg.SubmissionVariant(
            chromosomeCoordinates=chromosome_coordinates,
            copyNumber=self.copy_number,
            gene=gene,
            hgvs=self.hgvs,
            referenceCopyNumber=self.reference_copy_number,
            variantType=self.variant_type,
        )


@attrs.define(frozen=True)
class SubmissionVariantSet:
    variant: typing.List[SubmissionVariant]

    def to_msg(self) -> msg.SubmissionVariantSet:
        return msg.SubmissionVariantSet(
            variant=[msg_variant.to_msg() for msg_variant in self.variant]
        )


@attrs.define(frozen=True)
class SubmissionPhaseUnknownSet:
    hgvs: str
    variants: typing.List[SubmissionVariant]

    def to_msg(self) -> msg.SubmissionPhaseUnknownSet:
        return msg.SubmissionPhaseUnknownSet(
            hgvs=self.hgvs,
            variants=[msg_variant.to_msg() for msg_variant in self.variants],
        )


@attrs.define(frozen=True)
class SubmissionClinicalFeature:
    clinical_features_affected_status: ClinicalFeaturesAffectedStatus
    db: typing.Optional[ClinicalFeaturesDb] = None
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionClinicalFeature:
        return msg.SubmissionClinicalFeature(
            clinicalFeaturesAffectedStatus=self.clinical_features_affected_status,
            db=self.db,
            id=self.id,
            name=self.name,
        )


@attrs.define(frozen=True)
class SubmissionObservedIn:
    affected_status: AffectedStatus
    allele_origin: AlleleOrigin
    collection_method: CollectionMethod
    clinical_features: typing.Optional[typing.List[SubmissionClinicalFeature]] = None
    clinical_features_comment: typing.Optional[str] = None
    number_of_individuals: typing.Optional[int] = None
    struct_var_method_type: typing.Optional[StructVarMethodType] = None

    def to_msg(self) -> msg.SubmissionObservedIn:
        clinical_features = None
        if self.clinical_features:
            clinical_features = [
                SubmissionClinicalFeature.to_msg(msg_feature)
                for msg_feature in self.clinical_features
            ]
        return msg.SubmissionObservedIn(
            affectedStatus=self.affected_status,
            alleleOrigin=self.allele_origin,
            collectionMethod=self.collection_method,
            clinicalFeatures=clinical_features,
            clinicalFeaturesComment=self.clinical_features_comment,
            numberOfIndividuals=self.number_of_individuals,
            structVarMethodType=self.struct_var_method_type,
        )


@attrs.define(frozen=True)
class SubmissionHaplotypeSet:
    hgvs: str
    variants: typing.List[SubmissionVariant]
    star_allele_name: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionHaplotypeSet:
        return msg.SubmissionHaplotypeSet(
            hgvs=self.hgvs,
            variants=[msg_variant.to_msg() for msg_variant in self.variants],
            starAlleleName=self.star_allele_name,
        )


@attrs.define(frozen=True)
class SubmissionDistinctChromosomesSet:
    hgvs: str
    #: Hast at least two elements
    variants: typing.List[SubmissionVariant]

    def to_msg(self) -> msg.SubmissionDistinctChromosomesSet:
        return msg.SubmissionDistinctChromosomesSet(
            hgvs=self.hgvs,
            variants=[msg_variant.to_msg() for msg_variant in self.variants],
        )


@attrs.define(frozen=True)
class SubmissionHaplotypeSets:
    haplotype_set: typing.Optional[SubmissionHaplotypeSet] = None
    haplotype_single_variant_set: typing.Optional[SubmissionHaplotypeSet] = None

    def to_msg(self) -> msg.SubmissionHaplotypeSets:
        haplotype_set = None
        if self.haplotype_set:
            haplotype_set = self.haplotype_set.to_msg()
        haplotype_single_variant_set = None
        if self.haplotype_single_variant_set:
            haplotype_single_variant_set = self.haplotype_single_variant_set.to_msg()
        return msg.SubmissionHaplotypeSets(
            haplotypeSet=haplotype_set,
            haplotypeSingleVariantSet=haplotype_single_variant_set,
        )


@attrs.define(frozen=True)
class SubmissionDiplotypeSet:
    haplotype_sets: typing.List[SubmissionHaplotypeSets]
    hgvs: str
    star_allele_name: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionDiplotypeSet:
        return msg.SubmissionDiplotypeSet(
            haplotypeSets=[msg_sets.to_msg() for msg_sets in self.haplotype_sets],
            hgvs=self.hgvs,
            starAlleleName=self.star_allele_name,
        )


@attrs.define(frozen=True)
class SubmissionCitation:
    db: typing.Optional[CitationDb] = None
    id: typing.Optional[str] = None
    url: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionCitation:
        return msg.SubmissionCitation(
            db=self.db,
            id=self.id,
            url=self.url,
        )


@attrs.define(frozen=True)
class SubmissionAssertionCriteria:
    db: typing.Optional[CitationDb] = None
    id: typing.Optional[str] = None
    url: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionAssertionCriteria:
        return msg.SubmissionAssertionCriteria(
            db=self.db,
            id=self.id,
            url=self.url,
        )


@attrs.define(frozen=True)
class SubmissionCondition:
    db: typing.Optional[ConditionDb] = None
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionCondition:
        return msg.SubmissionCondition(
            db=self.db,
            id=self.id,
            name=self.name,
        )


@attrs.define(frozen=True)
class SubmissionDrugResponse:
    db: typing.Optional[ConditionDb] = None
    drug_name: typing.Optional[str] = None
    id: typing.Optional[str] = None
    condition: typing.Optional[typing.List[SubmissionCondition]] = None

    def to_msg(self) -> msg.SubmissionDrugResponse:
        condition = None
        if self.condition:
            condition = [msg_condition.to_msg() for msg_condition in self.condition]
        return msg.SubmissionDrugResponse(
            db=self.db,
            drugName=self.drug_name,
            id=self.id,
            condition=condition,
        )


@attrs.define(frozen=True)
class SubmissionConditionSet:
    condition: typing.Optional[typing.List[SubmissionCondition]] = None
    drug_response: typing.Optional[typing.List[SubmissionDrugResponse]] = None

    def to_msg(self) -> msg.SubmissionConditionSet:
        condition = None
        if self.condition:
            condition = [msg_condition.to_msg() for msg_condition in self.condition]
        drug_response = None
        if self.drug_response:
            drug_response = [msg_response.to_msg() for msg_response in self.drug_response]
        return msg.SubmissionConditionSet(
            condition=condition,
            drugResponse=drug_response,
        )


@attrs.define(frozen=True)
class SubmissionCompoundHeterozygoteSetVariantSet:
    variant_set: typing.Optional[SubmissionVariantSet] = None

    def to_msg(self) -> msg.SubmissionCompoundHeterozygoteSetVariantSet:
        variant_set = None
        if self.variant_set:
            variant_set = self.variant_set.to_msg()
        return msg.SubmissionCompoundHeterozygoteSetVariantSet(variantSet=variant_set)


@attrs.define(frozen=True)
class SubmissionCompoundHeterozygoteSet:
    hgvs: str
    # Must have two entries
    variant_sets: typing.List[SubmissionCompoundHeterozygoteSetVariantSet]

    def to_msg(self) -> msg.SubmissionCompoundHeterozygoteSet:
        return msg.SubmissionCompoundHeterozygoteSet(
            hgvs=self.hgvs,
            variantSets=[msg_set.to_msg() for msg_set in self.variant_sets],
        )


@attrs.define(frozen=True)
class SubmissionClinicalSignificance:
    clinical_significance_description: ClinicalSignificanceDescription
    citation: typing.Optional[typing.List[SubmissionCitation]] = None
    comment: typing.Optional[str] = None
    custom_assertion_score: typing.Optional[float] = None
    date_last_evaluated: typing.Optional[str] = None
    explanation_of_drug_response: typing.Optional[str] = None
    explanation_of_other_clinical_significance: typing.Optional[str] = None
    mode_of_inheritance: typing.Optional[ModeOfInheritance] = None

    def to_msg(self) -> msg.SubmissionClinicalSignificance:
        citation = None
        if self.citation:
            citation = [msg_citation.to_msg() for msg_citation in self.citation]
        return msg.SubmissionClinicalSignificance(
            clinicalSignificanceDescription=self.clinical_significance_description,
            citation=citation,
            comment=self.comment,
            customAssertionScore=self.custom_assertion_score,
            dateLastEvaluated=self.date_last_evaluated,
            explanationOfDrugResponse=self.explanation_of_drug_response,
            explanationOfOtherClinicalSignificance=self.explanation_of_other_clinical_significance,
            modeOfInheritance=self.mode_of_inheritance,
        )


@attrs.define(frozen=True)
class SubmissionClinvarSubmission:
    clinical_significance: SubmissionClinicalSignificance
    condition_set: SubmissionConditionSet
    observed_in: typing.List[SubmissionObservedIn]
    record_status: RecordStatus
    clinvar_accession: typing.Optional[str] = None
    compound_heterozygote_set: typing.Optional[SubmissionCompoundHeterozygoteSet] = None
    diplotype_set: typing.Optional[SubmissionDiplotypeSet] = None
    distinct_chromosomes_set: typing.Optional[SubmissionDistinctChromosomesSet] = None
    #: Has at least two elements in `variants`
    haplotype_set: typing.Optional[SubmissionHaplotypeSet] = None
    #: Has exactly one elements in `variants`
    haplotype_single_variant_set: typing.Optional[SubmissionHaplotypeSet] = None
    local_id: typing.Optional[str] = None
    local_key: typing.Optional[str] = None
    phase_unknown_set: typing.Optional[SubmissionPhaseUnknownSet] = None
    variant_set: typing.Optional[SubmissionVariantSet] = None
    #: Additional information from import.  Will not be used for conversion to message but can be converted back to
    #: external formats.
    extra_data: typing.Optional[typing.Dict[str, typing.Any]] = None

    def to_msg(self) -> msg.SubmissionClinvarSubmission:
        compound_heterozygote_set = None
        if self.compound_heterozygote_set:
            compound_heterozygote_set = self.compound_heterozygote_set.to_msg()
        diplotype_set = None
        if self.diplotype_set:
            diplotype_set = self.diplotype_set.to_msg()
        distinct_chromosomes_set = None
        if self.distinct_chromosomes_set:
            distinct_chromosomes_set = self.distinct_chromosomes_set.to_msg()
        haplotype_set = None
        if self.haplotype_set:
            haplotype_set = self.haplotype_set.to_msg()
        haplotype_single_variant_set = None
        if self.haplotype_single_variant_set:
            haplotype_single_variant_set = self.haplotype_single_variant_set.to_msg()
        phase_unknown_set = None
        if self.phase_unknown_set:
            phase_unknown_set = self.phase_unknown_set.to_msg()
        variant_set = None
        if self.variant_set:
            variant_set = self.variant_set.to_msg()
        return msg.SubmissionClinvarSubmission(
            clinicalSignificance=self.clinical_significance.to_msg(),
            conditionSet=self.condition_set.to_msg(),
            observedIn=[msg_observed_in.to_msg() for msg_observed_in in self.observed_in],
            recordStatus=self.record_status,
            clinvarAccession=self.clinvar_accession,
            compoundHeterozygoteSet=compound_heterozygote_set,
            diplotypeSet=diplotype_set,
            distinctChromosomesSet=distinct_chromosomes_set,
            haplotypeSet=haplotype_set,
            haplotypeSingleVariantSet=haplotype_single_variant_set,
            localID=self.local_id,
            localKey=self.local_key,
            phaseUnknownSet=phase_unknown_set,
            variantSet=variant_set,
        )


@attrs.define(frozen=True)
class SubmissionContainer:
    assertion_criteria: typing.Optional[SubmissionAssertionCriteria] = None
    behalf_org_id: typing.Optional[int] = None
    clinvar_deletion: typing.Optional[SubmissionClinvarDeletion] = None
    clinvar_submission: typing.Optional[typing.List[SubmissionClinvarSubmission]] = None
    clinvar_submission_release_status: typing.Optional[ReleaseStatus] = None
    submission_name: typing.Optional[str] = None

    def to_msg(self) -> msg.SubmissionContainer:
        assertion_criteria = None
        if self.assertion_criteria:
            assertion_criteria = self.assertion_criteria.to_msg()
        clinvar_deletion = None
        if self.clinvar_deletion:
            clinvar_deletion = self.clinvar_deletion.to_msg()
        clinvar_submission = None
        if self.clinvar_submission:
            clinvar_submission = [
                msg_submission.to_msg() for msg_submission in self.clinvar_submission
            ]
        return msg.SubmissionContainer(
            assertionCriteria=assertion_criteria,
            behalfOrgID=self.behalf_org_id,
            clinvarDeletion=clinvar_deletion,
            clinvarSubmission=clinvar_submission,
            clinvarSubmissionReleaseStatus=self.clinvar_submission_release_status,
            submissionName=self.submission_name,
        )



"""Data structures for the internal "business logic" representation.

These data structure use the ``snake_case`` used throughout the program.
"""

from clinvar_api.models.query_response import *  # noqa: F401, F403
from clinvar_api.models.sub_payload import *  # noqa: F401, F403
from clinvar_api.models.sub_response import *  # noqa: F401, F403



import datetime
import typing
import uuid

import cattr
import dateutil.parser


def _setup_converter() -> cattr.Converter:
    """Setup ``cattr`` converter for UUID and datetime."""
    result = cattr.Converter()
    result.register_structure_hook(uuid.UUID, lambda d, _: uuid.UUID(d))
    result.register_unstructure_hook(uuid.UUID, str)
    result.register_structure_hook(datetime.datetime, lambda d, _: dateutil.parser.parse(d))
    result.register_unstructure_hook(
        datetime.datetime,
        lambda obj: obj.replace(tzinfo=datetime.timezone.utc)
        .astimezone()
        .replace(microsecond=0)
        .isoformat(),
    )
    return result


#: cattr Converter to use
CONVERTER = _setup_converter()


def clean_for_json(
    value: typing.Union[
        bool, int, float, typing.List[typing.Any], None, typing.Dict[str, typing.Any]
    ]
) -> typing.Union[bool, int, float, typing.List[typing.Any], None, typing.Dict[str, typing.Any]]:
    """Clean the given value for JSON submission."""
    if isinstance(value, dict):
        return {k: clean_for_json(v) for k, v in value.items() if v is not None}
    elif isinstance(value, list):
        return [clean_for_json(elem) for elem in value]
    else:
        return value



"""Extract variants from RCV records."""

import contextlib
import enum
import gzip
import json
import os
import typing

import attrs
import cattrs
from logzero import logger
import tqdm

from clinvar_data import models
from clinvar_data.cattrs_helpers import CONVERTER
from clinvar_data.models import MeasureAttributeType


class OutputFilesHandler:
    def __init__(self, output_dir: str, gzip_output: bool):
        self.output_dir = output_dir
        self.gzip_output = gzip_output
        self.suffix = "tsv.gz" if self.gzip_output else "tsv"
        self.files: typing.Dict[typing.Tuple[str, str], typing.TextIO] = {}

    def get_file(self, stack: contextlib.ExitStack, assembly: str, variant_size: str):
        key = (assembly, variant_size)
        if key not in self.files:
            file_path = os.path.join(
                self.output_dir, f"clinvar-variants-{assembly}-{variant_size}.{self.suffix}"
            )
            if self.gzip_output:
                self.files[key] = gzip.open(file_path, "wt")
            else:
                self.files[key] = open(file_path, "wt")
            stack.push(self.files[key])
        return self.files[key]


@enum.unique
class VariantType(enum.Enum):
    INSERTION = "insertion"
    DELETION = "deletion"
    SNV = "single nucleotide variant"
    INDEL = "indel"
    DUPLICATION = "duplication"
    TANDEM_DUPLICATION = "tandem duplication"
    STRUCTURAL_VARIANT = "structural variant"
    COPY_NUMBER_GAIN = "copy number gain"
    COPY_NUMBER_LOSS = "copy number loss"
    PROTEIN_ONLY = "protein only"
    MICROSATELLITE = "microsatellite"
    INVERSION = "inversion"
    OTHER = "other"

    @classmethod
    def from_measure_type(cls, mt: typing.Optional[models.MeasureType]) -> "VariantType":
        mapping: typing.Dict[typing.Optional[models.MeasureType], VariantType] = {
            models.MeasureType.INSERTION: VariantType.INSERTION,
            models.MeasureType.DELETION: VariantType.DELETION,
            models.MeasureType.SNV: VariantType.SNV,
            models.MeasureType.INDEL: VariantType.INDEL,
            models.MeasureType.DUPLICATION: VariantType.DUPLICATION,
            models.MeasureType.TANDEM_DUPLICATION: VariantType.TANDEM_DUPLICATION,
            models.MeasureType.STRUCTURAL_VARIANT: VariantType.STRUCTURAL_VARIANT,
            models.MeasureType.COPY_NUMBER_GAIN: VariantType.COPY_NUMBER_GAIN,
            models.MeasureType.COPY_NUMBER_LOSS: VariantType.COPY_NUMBER_LOSS,
            models.MeasureType.INVERSION: VariantType.INVERSION,
        }
        return mapping.get(mt, VariantType.OTHER)


@attrs.frozen(auto_attribs=True)
class VariantRecord:
    rcv: str
    variant_type: VariantType
    sequence_location: models.SequenceLocation
    absolute_copy_number: typing.Union[None, str, int, float] = None
    reference_copy_number: typing.Union[None, str, int, float] = None
    copy_number_tuple: typing.Union[None, str, int, float] = None


def run(path_input: str, output_dir: str, gzip_output: bool):
    os.makedirs(output_dir, exist_ok=True)

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    output_files = OutputFilesHandler(output_dir, gzip_output)

    with inputf, contextlib.ExitStack() as stack:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            dict_value = json.loads(line)
            clinvar_set = CONVERTER.structure(dict_value, models.ClinVarSet)
            rca = clinvar_set.reference_clinvar_assertion

            if rca.measures:
                for measure in rca.measures.measures or []:
                    # print(measure.sequence_locations)
                    if not measure.sequence_locations:
                        continue

                    absolute_copy_number = None
                    reference_copy_number = None
                    copy_number_tuple = None
                    for attribute in measure.attributes or []:
                        value = attribute.value or attribute.integer_value
                        if attribute.type == MeasureAttributeType.ABSOLUTE_COPY_NUMBER:
                            absolute_copy_number = value
                        elif attribute.type == MeasureAttributeType.REFERENCE_COPY_NUMBER:
                            reference_copy_number = value
                        elif attribute.type == MeasureAttributeType.COPY_NUMBER_TUPLE:
                            copy_number_tuple = value

                    # The structural variants that we saw had the sequence location directly on the measure.
                    for sequence_location in measure.sequence_locations:
                        logger.info("%s", sequence_location)
                        record = VariantRecord(
                            rcv=rca.clinvar_accession.acc,
                            variant_type=VariantType.from_measure_type(measure.type),
                            sequence_location=sequence_location,
                            absolute_copy_number=absolute_copy_number,
                            reference_copy_number=reference_copy_number,
                            copy_number_tuple=copy_number_tuple,
                        )
                        if record.variant_type == VariantType.OTHER:
                            continue

                        if (
                            sequence_location.reference_allele
                            or sequence_location.alternate_allele
                            or sequence_location.reference_allele_vcf
                            or sequence_location.alternate_allele_vcf
                        ):
                            variant_size = "seqvars"
                        else:
                            variant_size = "strucvars"
                        dest = output_files.get_file(
                            stack, sequence_location.assembly.lower(), variant_size
                        )
                        print(cattrs.unstructure(record), file=dest)



"""Code to convert ClinVar XML to JSONL"""

import datetime
import gzip
import json
import sys
import traceback
import typing

import cattrs
import click
import tqdm
import xmltodict

from clinvar_data import models


def remove_empties_from_containers(
    container: typing.Union[dict, list]
) -> typing.Union[dict, list, None]:
    if isinstance(container, list):
        new_list = []
        for v in container:
            if isinstance(v, (dict, list)):
                v = remove_empties_from_containers(v)
            if v is not None:
                new_list.append(v)
        return new_list or None
    elif isinstance(container, dict):
        new_dict = {}
        for k, v in container.items():
            if isinstance(v, (dict, list)):
                v = remove_empties_from_containers(v)
            if v is not None:
                new_dict[k] = v
        return new_dict or None
    else:  # pragma: no cover
        assert False, "must not happen"


# Define a custom function to serialize datetime objects
def json_default(obj):
    if isinstance(obj, datetime.datetime):
        return obj.strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")
    else:
        raise TypeError("Type not serializable")


def convert_clinvarset(json_cv: dict) -> models.ClinVarSet:
    """Convert a ClinVarSet from JSON dict."""
    return models.ClinVarSet.from_json_data(json_cv)


def convert(
    input_file: str, output_file: str, max_records: int = 0, use_click: bool = False
) -> int:
    """Run conversion from ClinVar XML to JSONL"""
    if input_file.endswith((".gz", ".bgz")):
        inputf: typing.Union[typing.BinaryIO, gzip.GzipFile] = gzip.open(input_file, "rb")
    elif use_click:
        inputf = click.open_file(input_file, "rb")
    else:
        inputf = open(input_file, "rb")

    if output_file.endswith(".gz"):
        outputf = gzip.open(output_file, "wt")
    elif use_click:
        outputf = click.open_file(output_file, "wt")
    else:
        outputf = open(output_file, "wt")

    pb = tqdm.tqdm(desc="parsing", unit=" ClinVarSet records", smoothing=1.0)
    records_written = 0
    errors = 0

    def handle_clinvarset(_, json_cvs: dict):
        """Handle single ClinVarSet entry after parsing by ``xmltodict``."""
        try:
            data = convert_clinvarset(json_cvs)
        except Exception:  # pragma: no cover
            nonlocal errors
            errors += 1
            print("Problem with data: exception and data follow", file=sys.stderr)
            traceback.print_exc()
            print(json_cvs, file=sys.stderr)
            return True

        print(
            json.dumps(
                remove_empties_from_containers(cattrs.unstructure(data)), default=json_default
            ),
            file=outputf,
        )

        nonlocal records_written
        records_written += 1
        pb.update(1)

        return max_records == 0 or records_written < max_records

    try:
        xmltodict.parse(inputf, item_depth=2, item_callback=handle_clinvarset)
    except xmltodict.ParsingInterrupted:  # pragma: no cover
        print(f"stopping after parsing {records_written} records", file=sys.stderr)

    if errors == 0:
        return 0
    else:
        print(f"a total of {errors} errors occurred", file=sys.stderr)
        return 1



"""REST API client code for communicating with server endpoints."""

import json
import typing

import attrs
import cattrs
from jsonschema import ValidationError
from logzero import logger
import requests

from clinvar_api import common, exceptions, models, msg, schemas

#: URL of the server endpoint (non-test/production).
ENDPOINT_URL_PROD = "https://submit.ncbi.nlm.nih.gov/api/v1/submissions/"

#: URL of the test endpoint.
ENDPOINT_URL_TEST = "https://submit.ncbi.nlm.nih.gov/apitest/v1/submissions/"

#: URL suffix for enabling dry-run.
SUFFIX_DRYRUN = "?dry-run=true"


def _obfuscate_repr(s):
    """Helper function for obfustating passwords"""
    if len(s) < 5:
        return repr("*" * len(s))
    else:
        return repr(s[:5] + "*" * (len(s) - 5))


@attrs.define(frozen=True)
class Config:
    """Configuration for the ``Client`` class."""

    #: Token to use for authentication.
    auth_token: str = attrs.field(repr=_obfuscate_repr)

    #: Whether to use the test endpoint.
    use_testing: bool = False

    #: Whether to use dry running.
    use_dryrun: bool = False

    #: Whether to validate submission payload before posting.
    presubmission_validation: bool = True

    #: Whether or not to verify SSL on submission.
    verify_ssl: bool = True


def submit_data(submission_container: models.SubmissionContainer, config: Config) -> models.Created:
    """Submit new data to ClinVar API.

    :param payload: The submission data.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.SubmissionFailed: on problems with the submission.
    """
    logger.info("Submitting with config %s", config)

    url_prefix = ENDPOINT_URL_TEST if config.use_testing else ENDPOINT_URL_PROD
    url_suffix = SUFFIX_DRYRUN if config.use_dryrun else ""
    url = f"{url_prefix}{url_suffix}"
    logger.debug("Will submit to URL %s", url)
    headers = {
        "SP-API-KEY": config.auth_token,
    }

    payload = cattrs.unstructure(submission_container.to_msg())
    logger.debug("Payload data is %s", json.dumps(payload, indent=2))
    cleaned_payload = common.clean_for_json(payload)
    logger.debug("Cleaned payload data is %s", json.dumps(cleaned_payload, indent=2))
    if config.presubmission_validation:
        logger.info("Validating payload...")
        schemas.validate_submission_payload(cleaned_payload)
        logger.info("... done validating payload")
    else:
        logger.info("Configured to NOT validate payload before submission")

    post_data = {
        "actions": [
            {"type": "AddData", "targetDb": "clinvar", "data": {"content": cleaned_payload}}
        ]
    }
    logger.debug("Overall POST payload is %s", post_data)

    response = requests.post(url, headers=headers, json=post_data, verify=config.verify_ssl)

    if response.ok:
        logger.info("API returned OK - %s:  %s", response.status_code, response.reason)
        if response.status_code == 204:  # no content, on dry-run
            logger.info("Server returned '204: No Content', constructing fake created message.")
            return models.Created(id="--NONE--dry-run-result--")
        else:
            created_msg = common.CONVERTER.structure(response.json(), msg.Created)
            return models.Created.from_msg(created_msg)
    else:
        logger.warning("API returned an error - %s: %s", response.status_code, response.reason)
        error_msg = common.CONVERTER.structure(response.json(), msg.Error)
        error_obj = models.Error.from_msg(error_msg)
        logger.debug("Full server response is %s", response.json())
        if hasattr(error_obj, "errors"):
            raise exceptions.SubmissionFailed(
                f"ClinVar submission failed: {error_obj.message}, errors: {error_obj.errors}"
            )
        else:
            raise exceptions.SubmissionFailed(f"ClinVar submission failed: {error_obj.message}")


@attrs.define(frozen=True)
class RetrieveStatusResult:
    """Result type for ``retrieve_status`` function."""

    #: The submission status.
    status: models.SubmissionStatus
    #: A dict mapping file URLs to the parsed ``Sum``.
    summaries: typing.Dict[str, models.SummaryResponse]


def _retrieve_status_summary(
    url: str, validate_response_json: bool = True
) -> models.SummaryResponse:
    """Retrieve status summary from the given URL."""
    response = requests.get(url)
    if response.ok:
        response_json = response.json()
        if validate_response_json:
            logger.debug("Validating status summary response ...")
            try:
                schemas.validate_status_summary(response_json)
            except ValidationError as e:
                logger.warning("Response summary validation JSON is invalid: %s", e)
            logger.debug("... done validating status summary response")
        sr_msg = cattrs.structure(response.json(), msg.SummaryResponse)
        return models.SummaryResponse.from_msg(sr_msg)
    else:
        raise exceptions.QueryFailed(
            f"Could not perform query: {response.status_code} {response.reason}"
        )


def retrieve_status(
    submission_id: str,
    config: Config,
) -> RetrieveStatusResult:
    """Retrieve submission status from API.

    :param submission_id: The identifier of the submission as returned earlier from API.
    :param config: The connfiguration to use.
    :return: The information about the created submission.
    :raises exceptions.QueryFailed: on problems with the communication to the server.
    """
    url_prefix = ENDPOINT_URL_TEST if config.use_testing else ENDPOINT_URL_PROD
    url_suffix = SUFFIX_DRYRUN if config.use_dryrun else ""
    url = f"{url_prefix}{submission_id}/actions/{url_suffix}"
    headers = {
        "SP-API-KEY": config.auth_token,
    }
    logger.debug("Will query URL %s", url)
    response = requests.get(url, headers=headers)
    if response.ok:
        logger.info("API returned OK - %s: %s", response.status_code, response.reason)
        logger.debug("Structuring response ...")
        status_msg = common.CONVERTER.structure(response.json(), msg.SubmissionStatus)
        logger.debug(
            "structured response is %s",
            json.dumps(common.CONVERTER.unstructure(status_msg), indent=2),
        )
        logger.debug("... done structuring response")
        status_obj = models.SubmissionStatus.from_msg(status_msg)
        logger.info(
            "Attempting to fetch %d status summary files...",
            len(
                [
                    None
                    for action in status_obj.actions
                    for action_response in action.responses
                    for _ in action_response.files
                ]
            ),
        )
        summaries = {}
        for action in status_obj.actions:
            for action_response in action.responses:
                for file_ in action_response.files:
                    logger.info(" - fetching %s", file_.url)
                    summaries[file_.url] = _retrieve_status_summary(file_.url)
        logger.info("... done fetching status summary files")
        return RetrieveStatusResult(status=status_obj, summaries=summaries)
    else:
        logger.info("API returned an error %s: %s", response.status_code, response.reason)
        response_json = response.json()
        raise exceptions.QueryFailed(f"ClinVar query failed: {response_json}")


class Client:
    """NCBI ClinVar REST API client."""

    def __init__(self, config: Config):
        self.config = config

    def submit_data(self, payload: models.SubmissionContainer) -> models.Created:
        """Submit new data to ClinVar API.

        :param payload: The submission data.
        :return: The information about the created submission.
        :raises exceptions.SubmissionFailed: on problems with the submission.
        """
        return submit_data(payload, self.config)

    def retrieve_status(self, submission_id: str) -> RetrieveStatusResult:
        """Retrieve submission status from API.

        :param submission_id: The identifier of the submission as returned earlier from API.
        :return: The information about the created submission.
        :raises exceptions.QueryFailed: on problems with the communication to the server.
        """
        return retrieve_status(submission_id, self.config)



"""Data structures used for communicating with the server endpoints.

These data structures use the same case (``camelCaseWithIDCase``) as the ClinVar API JSON.
"""

from clinvar_api.msg.query_response import *  # noqa: F401, F403
from clinvar_api.msg.sub_payload import *  # noqa: F401, F403
from clinvar_api.msg.sub_response import *  # noqa: F401, F403



"""Console script for ClinVar This!"""

import typing

import attrs
import click

from clinvar_data import (
    class_by_freq,
    conversion,
    extract_vars,
    gene_impact,
    phenotype_link,
)
from clinvar_this import _version, batches, exceptions
from clinvar_this.config import Config, dump_config, load_config, save_config


@click.group()
@click.version_option(_version.__version__)
@click.option("--verbose/--no-verbose", default=False)
@click.option("--profile", default="default", help="The profile to use")
@click.option(
    "--verify-ssl/--no-verify-ssl", default=True, help="Whether to enable SSL verification"
)
@click.pass_context
def cli(ctx: click.Context, verbose: bool, profile: str, verify_ssl: bool):
    """Main entry point for CLI via click."""
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    ctx.obj["profile"] = profile
    ctx.obj["verify_ssl"] = verify_ssl


@cli.group()
def config():
    """Sub command category ``varfish-this config ...``"""


@config.command("set")
@click.argument("name")
@click.argument("value")
@click.pass_context
def config_set(ctx: click.Context, name: str, value: str):
    """Sub command ``varfish-this config set NAME VALUE``

    Set the configuration variable with the given ``NAME`` to the given ``VALUE``.  This will interpret the
    current ``--profile`` setting.
    """
    profile: str = ctx.obj["profile"]
    try:
        config_obj = load_config(profile)
    except exceptions.ConfigFileMissingException:
        config_obj = Config(profile=profile, auth_token="")  # swallow, will recreate
    allowed_names = ["auth_token"]
    if name not in allowed_names:
        raise click.ClickException(f"Invalid value {name}, must be one of {allowed_names}")
    # We need to ignore type checking in the following line because of a false positive:
    #
    # error: Argument 2 to "evolve" of "Config" has incompatible type "**dict[str, str]";
    # expected "bool"  [arg-type]
    config_obj = attrs.evolve(config_obj, **{name: value})  # type: ignore
    save_config(config_obj, profile)


@config.command("get")
@click.option("--profile", default="default", help="The profile to get the value from")
@click.argument("name")
def config_get(profile: str, name: str):
    """Sub command ``varfish-this config get NAME``

    Show the configuration variable with the given ``NAME``.  This will interpret the current ``--profile`` setting.
    """
    config = load_config(profile)
    print(getattr(config, name, "<undefined>"))


@config.command("dump")
def config_dump():
    """Sub command ``varfish-this config dump``

    Print the configuration file to stdout.
    """
    dump_config()


@cli.group("batch")
def batch():
    """Sub comment category ``batch ...``"""


@batch.command("list")
@click.pass_context
def batch_list(ctx: click.Context):
    """List existing batches"""
    config_obj = load_config(ctx.obj["profile"])
    batches.list_(config_obj)


@batch.command("import")
@click.argument("path")
@click.option(
    "--metadata",
    "-m",
    required=False,
    multiple=True,
    help="Provide meta data settings as KEY=VALUE settings",
)
@click.option("--name", required=False, default=None, help="Name of the batch to create or add to")
@click.pass_context
def batch_import(
    ctx: click.Context,
    path: str,
    name: typing.Optional[str] = None,
    metadata: typing.Optional[typing.Tuple[str, ...]] = None,
):
    """Import data for a new or existing batch"""
    config_obj = load_config(ctx.obj["profile"])
    if not name:
        name = batches.gen_name(config_obj)
        print(f"Using name = {name}")
    if not metadata:
        metadata = ()
    print(f"metadata = {metadata}")
    config_obj = load_config(ctx.obj["profile"])
    batches.import_(config_obj, name, path, metadata)


@batch.command("export")
@click.argument("name")
@click.argument("path")
@click.option("--force/--no-force", required=False, default=False, help="Overwrite existing files")
@click.option(
    "--struc-var/--no-struc-var",
    required=False,
    default=False,
    help="Export structural variants rather than sequence variants",
)
@click.pass_context
def batch_export(
    ctx: click.Context,
    name: str,
    path: str,
    force: bool = False,
    struc_var: bool = False,
):
    """Export batch data to a given file"""
    config_obj = load_config(ctx.obj["profile"])
    batches.export(config_obj, name, path, force, struc_var)


@batch.command("update-metadata")
@click.argument("name")
@click.argument("metadata", nargs=-1)
@click.pass_context
def batch_update_metadata(
    ctx: click.Context, name: str, metadata: typing.Optional[typing.Tuple[str, ...]] = None
):
    """Update batch metadata without importing files"""
    if not metadata:
        metadata = ()
    print(metadata)
    config_obj = load_config(ctx.obj["profile"])
    batches.update_metadata(config_obj, name, metadata)


@batch.command("submit")
@click.option(
    "--use-testing/--no-testing",
    required=False,
    default=False,
    help="Whether to use the testing API",
)
@click.option(
    "--dry-run/--no-dry-run",
    required=False,
    default=False,
    help="Whether to use the ClinVar dry-run",
)
@click.argument("name")
@click.pass_context
def batch_submit(ctx: click.Context, use_testing: bool, dry_run: bool, name: str):
    """Submit the given batch to ClinVar"""
    config_obj = load_config(ctx.obj["profile"])
    config_obj = attrs.evolve(config_obj, verify_ssl=ctx.obj["verify_ssl"])
    batches.submit(config_obj, name, use_testing=use_testing, dry_run=dry_run)


@batch.command("retrieve")
@click.option(
    "--use-testing/--no-testing",
    required=False,
    default=False,
    help="Whether to use the testing API",
)
@click.argument("name")
@click.pass_context
def batch_retrieve(ctx: click.Context, use_testing: bool, name: str):
    """Submit the given batch to ClinVar"""
    config_obj = load_config(ctx.obj["profile"])
    config_obj = attrs.evolve(config_obj, verify_ssl=ctx.obj["verify_ssl"])
    batches.retrieve(config_obj, name, use_testing=use_testing)


@cli.group("data")
def data():
    """Sub command category "data"."""


@data.command("xml-to-jsonl")
@click.argument("input_file")
@click.argument("output_file")
@click.option(
    "--max-records", required=False, default=0, help="Maximum number of records to convert"
)
@click.pass_context
def xml_to_jsonl(ctx: click.Context, input_file: str, output_file: str, max_records: int):
    """Convert XML to JSONL"""
    retcode = conversion.convert(input_file, output_file, max_records=max_records, use_click=True)
    ctx.exit(retcode)


@data.command("gene-variant-report")
@click.argument("input_file")
@click.argument("output_file")
@click.pass_context
def gene_impact_report(ctx: click.Context, input_file: str, output_file: str):
    """Create a gene variant summary report."""
    _ = ctx
    gene_impact.run_report(input_file, output_file)


@data.command("gene-phenotype-links")
@click.argument("input_file")
@click.argument("output_file")
@click.option(
    "--needs-hpo-terms/--no-needs-hpo-terms",
    type=bool,
    default=True,
    help="Whether to filter to rows with HPO terms (default: true)",
)
@click.pass_context
def gene_phenotype_links(
    ctx: click.Context, input_file: str, output_file: str, needs_hpo_terms: bool
):
    """Create links between gene and phenotype."""
    _ = ctx
    phenotype_link.run_report(input_file, output_file, needs_hpo_terms=needs_hpo_terms)


@data.command("acmg-class-by-freq")
@click.argument("input_file")
@click.argument("output_file")
@click.option(
    "--thresholds",
    type=str,
    default=",".join(map(str, class_by_freq.DEFAULT_THRESHOLDS)),
    help="Whether to filter to rows with HPO terms (default: true)",
)
@click.pass_context
def acmg_class_by_freq(ctx: click.Context, input_file: str, output_file: str, thresholds: str):
    """Create links between gene and phenotype."""
    _ = ctx
    thresholds_float = list(map(float, thresholds.split(",")))
    class_by_freq.run_report(input_file, output_file, thresholds=thresholds_float)


@data.command("extract-vars")
@click.argument("path_input")
@click.argument("path_output_dir")
@click.option(
    "--gzip-output/--no-gzip-output", default=True, help="Whether to gzip output (default: true)"
)
@click.pass_context
def cli_extract_vars(ctx: click.Context, path_input: str, path_output_dir: str, gzip_output: bool):
    """Write out variants from RCV records."""
    _ = ctx
    extract_vars.run(path_input, path_output_dir, gzip_output)



"""Exceptions used in ``clinvar_this`` module."""


class ClinvarThisException(Exception):
    """Base exception class."""


class IOException(ClinvarThisException):
    """Raised on problems with I/O in ``clinvar-this``."""


class ConfigException(ClinvarThisException):
    """Raised on configuration problems with ``clinvar-this``."""


class ConfigFileMissingException(ClinvarThisException):
    """Raised if the configuration file is missing."""


class ArgumentsError(ClinvarThisException):
    """Raised on problems with program arguments."""


class InvalidFormat(ClinvarThisException):
    """Raised on problems file contents in ``clinvar-this``."""
